/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// An IO class. Cause you know, it's a stream. In this particular case, it's only I
// though, no O in this class. Unless you count the fact that stuff gets returned, which
// is output, sorta. If you catch my drift.
package com.kirkwarez.kirkpack.io;

// We need IO stuff.
import java.io.*;

/**
 * <p>This stream behaves about exactly as <code>DataInputStream</code>, it reads
 * primitive and some non-primitive data types from the underlying stream. The reason it
 * doesn't extend <code>DataInputStream</code> is because all methods in that class are
 * declared final. This class does implement <code>DataInput</code>.</p>
 * 
 * <p>The big difference is that this class takes a byte order on construction, and
 * guarantees that the primitives will be constructed from bytes read using that byte
 * order.</p>
 * 
 * <p>The non-primitive methods from this class are delegated to a
 * <code>DataInputStream</code>, however the <code>readLine</code> method shouldn't be
 * used, because the <code>DataInputStream</code> method is deprecated, even though it's
 * implementing an interface... As of yet there is no replacement for this method, but
 * you shouldn't be using it anyways since lines are pretty rare in byte streams...</p>
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 */
public class EndianInputStream extends FilterInputStream implements DataInput {
	/**
	 * Constucts a new stream that delivers primitives read using the given byte order.
	 * 
	 * @param input The stream to read from.
	 * @param byteorder The byte order. This is either BIG_ENDIAN or LITTLE_ENDIAN.
	 */
	public EndianInputStream(InputStream input, int byteorder) {
		// Super the input stream.
		super(new DataInputStream(input));
		
		// Since we don't want to cast the entire freaking time, we cache the uncast
		// InputStream.
		cache = (DataInputStream) in;
		
		// Set the byteorder.
		setByteorder(byteorder);
	}
	
	
	/**
	 * Creates a byte array wih the given length, fills it in with bytes from the stream
	 * completely and returns it.
	 *
	 * @param length The size of the buffer where the data is to be read into.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 */
	public byte[] readFully(int length) throws IOException {
		// Create the byte array.
		byte[] buffer = new byte[length];
		
		// Overload.
		readFully(buffer);
		
		// Return the now filled byte array.
		return buffer;
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @param buffer The buffer where the data is read into.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readFully(byte[])
	 */
	public void readFully(byte buffer[]) throws IOException {
		// Delegate.
		cache.readFully(buffer);
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @param buffer The buffer where the data is read into.
	 * @param offset The starting offset of the data in the array.
	 * @param length The number of bytes read.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readFully(byte[], int, int)
	 */
	public void readFully(byte buffer[], int offset, int length) throws IOException {
		// Delegate.
		cache.readFully(buffer, offset, length);
	}
	
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @param amount The number of bytes to be skipped.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#skipBytes(int)
	 */
	public int skipBytes(int amount) throws IOException {
		// Delegate.
		return cache.skipBytes(amount);
	}
	
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next boolean in this stream.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readBoolean()
	 */
	public boolean readBoolean() throws IOException {
		// Delegate.
		return cache.readBoolean();
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next byte in this stream.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readByte()
	 */
	public byte readByte() throws IOException {
		// Delegate.
		return cache.readByte();
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next unsigned byte in this stream as an int.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readUnsignedByte()
	 */
	public int readUnsignedByte() throws IOException {
		// Delegate.
		return cache.readUnsignedByte();
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next short read from this stream using the set byte order.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readShort()
	 */
	public short readShort() throws IOException {
		// Read in two bytes.
		byte[] bytes = readFully(2);
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Create and return the short the big endian way.
			return (short)
				(
					((bytes[0] & 0xFF) << 8) |
					((bytes[1] & 0xFF)     ) 
				);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Create and return the short the little endian way.
			return (short)
				(
					((bytes[1] & 0xFF) << 8) |
					((bytes[0] & 0xFF)     )
				);
		}
		else  throw new InternalError("Unknown byte order.");
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next unsigned short as an int read from this stream using the set byte
	 * order.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readUnsignedShort()
	 */
	public int readUnsignedShort() throws IOException {
		// Read in two bytes.
		byte[] bytes = readFully(2);
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Create and return the unsigned short the big endian way.
			return
				(
					((bytes[0] & 0xFF) << 8) |
					((bytes[1] & 0xFF)     ) 
				);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Create and return the unsigned short the little endian way.
			return
				(
					((bytes[1] & 0xFF) << 8) |
					((bytes[0] & 0xFF)     )
				);
		}
		else  throw new InternalError("Unknown byte order.");
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next Unicode character read from this stream, by reading two bytes and
	 * combining them. Why anyone would want to write out Unicode in such a crude way I
	 * don't know. If you can, look for cooler options. At least it's not bothered by
	 * byte order :^P
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readChar()
	 */
	public char readChar() throws IOException {
		// Delegate.
		return cache.readChar();
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next int read from this stream using the set byte order.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readInt()
	 */
	public int readInt() throws IOException {
		// Read in four bytes.
		byte[] bytes = readFully(4);
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Create and return the int the big endian way.
			return
				(
					((bytes[0] & 0xFF) << 24) |
					((bytes[1] & 0xFF) << 16) |
					((bytes[2] & 0xFF) <<  8) |
					((bytes[3] & 0xFF)      ) 
				);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Create and return the int the little endian way.
			return
				(
					((bytes[3] & 0xFF) << 24) |
					((bytes[2] & 0xFF) << 16) |
					((bytes[1] & 0xFF) <<  8) |
					((bytes[0] & 0xFF)      ) 
				);
		}
		else  throw new InternalError("Unknown byte order.");
	}
	
	/**
	 * Reads 4 bytes and turns them into an unsigned int. Returns it as a long,
	 * naturally.
	 *
	 * @return The next unsigned int read from this stream using the set byte order.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 */
	public long readUnsignedInt() throws IOException {
		// Read in four bytes.
		byte[] bytes = readFully(4);
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Create and return the unsigned int the big endian way.
			return
				(
					((bytes[0] & 0xFFl) << 24) |
					((bytes[1] & 0xFFl) << 16) |
					((bytes[2] & 0xFFl) <<  8) |
					((bytes[3] & 0xFFl)      ) 
				);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Create and return the unsigned int the little endian way.
			return
				(
					((bytes[3] & 0xFFl) << 24) |
					((bytes[2] & 0xFFl) << 16) |
					((bytes[1] & 0xFFl) <<  8) |
					((bytes[0] & 0xFFl)      ) 
				);
		}
		else  throw new InternalError("Unknown byte order.");
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next long read from this stream using the set byte order.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readLong()
	 */
	public long readLong() throws IOException {
		// Read in eight whole bytes!
		byte[] bytes = readFully(8);
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Create and return the long the big endian way.
			return 
				(
					((bytes[0] & 0xFFl) << 56) |
					((bytes[1] & 0xFFl) << 48) |
					((bytes[2] & 0xFFl) << 40) |
					((bytes[3] & 0xFFl) << 32) |
					((bytes[4] & 0xFFl) << 24) |
					((bytes[5] & 0xFFl) << 16) |
					((bytes[6] & 0xFFl) <<  8) |
					((bytes[7] & 0xFFl)      ) 
				);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Create and return the long the little endian way.
			return 
				(
					((bytes[7] & 0xFFl) << 56) |
					((bytes[6] & 0xFFl) << 48) |
					((bytes[5] & 0xFFl) << 40) |
					((bytes[4] & 0xFFl) << 32) |
					((bytes[3] & 0xFFl) << 24) |
					((bytes[2] & 0xFFl) << 16) |
					((bytes[1] & 0xFFl) <<  8) |
					((bytes[0] & 0xFFl)      ) 
				);
		}
		else  throw new InternalError("Unknown byte order.");
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next float read from this stream using the set byte order.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readFloat()
	 * @see java.lang.Float#intBitsToFloat(int)
	 */
	public float readFloat() throws IOException {
		// Read an int and convert that.
		return Float.intBitsToFloat(readInt());
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @return The next double read from this stream using the set byte order.
	 * @throws EOFException When the end of the steam is encountered.
	 * @throws IOException When I/O errors happen.
	 * @see java.io.DataInput#readDouble()
	 * @see java.lang.Double#longBitsToDouble(long)
	 */
	public double readDouble() throws IOException {
		// Read a long and convert that into a double.
		return Double.longBitsToDouble(readLong());
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 *
	 * @deprecated <code>DataInputStream</code> uses a deprecated way. However, using a
	 * <code>BufferedReader</code> internally isn't going to work, since it'll mess up
	 * the stream (it buffers), and will require two objects each method call, since you
	 * can't close it. Nobody should actually use this method, because it simply doesn't
	 * make sense. A line in a byte stream?* A null terminated string I can understand...
	 * but a line? I mean, strings are usually just stored as either null-terminted
	 * strings or with some length attribute. Just read that out and bother with the
	 * lines later.
	 * @return The next line of text in this stream.
	 * @exception IOException When I/O errors occur.
	 * @see java.io.BufferedReader#readLine()
	 * @see java.io.DataInputStream#readLine()
	 * @see java.io.DataInput#readLine()
	 */
	public String readLine() throws IOException {
		// Delegate :P
		return cache.readLine();
	}
	
	/**
	 * See the general contract of <code>DataInput</code>.
	 * 
	 * @return The Unicode string read from the underlying stream.
	 * @throws EOFException When the end of the stream is encountered.
	 * @throws IOException When IO errors happen.
	 * @throws UTFDataFormatException When the bytes do not represent a valid UTF-8
	 * encoding of a string.
     * @see java.io.DataInput#readUTF()
     */
	public String readUTF() throws IOException {
		// Delegate.
		return cache.readUTF();
	}
	
	
	/**
	 * Returns the byte order.
	 * 
	 * @return The byte order used by this stream.
	 */
	public int getByteorder() {
		return byteorder;
	}
	
	/**
	 * Sets the byte order.
	 * 
	 * @param byteorder The byte order used by this stream. 
	 */
	protected void setByteorder(int byteorder) {
		// Check.
		if(byteorder == BIG_ENDIAN || byteorder == LITTLE_ENDIAN)  this.byteorder = byteorder;
		else  throw new IllegalArgumentException("Unsupported byte order.");
	}
	
	
	/**
	 * The <code>InputStream</code> that <code>FilterInputStream</code> parent class
	 * delegates to, only uncast.
	 */
	protected DataInputStream cache;
	
	/**
	 * The byte order used by this stream.
	 */
	protected int byteorder;
	
	
	/**
	 * The Big Endian byte order.
	 */
	public static final int BIG_ENDIAN = 0;
	
	/**
	 * The Little Endian byte order.
	 */
	public static final int LITTLE_ENDIAN = 1;
}