/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// An IO class. Cause you know, it's a stream. In this particular case, it's only O
// though, no I in this class. Unless you count the fact that stuff is given to this
// class, which is output, sorta. If you catch my drift.
package com.kirkwarez.kirkpack.io;

// We need IO stuff.
import java.io.*;

/**
 * <p>This stream behaves about exactly as <code>DataOutputStream</code>, it writes
 * primitive and some non-primitive data types to the underlying stream. The reason it
 * doesn't extend <code>DataOutputStream</code> is because all methods in that class are
 * declared final. This class does implement <code>DataOutput</code>.</p>
 * 
 * <p>The big difference is that this class takes a byte order on construction, and
 * guarantees that the primitives' bytes will be written out using that byte
 * order.</p>
 * 
 * <p>The non-primitive methods from this class are delegated to a
 * <code>DataOutputStream</code>.</p>
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 */
public class EndianOutputStream extends FilterOutputStream implements DataOutput {
	/**
	 * Constucts a new stream that writes out primitives using the given byte order.
	 * 
	 * @param output The stream to write to.
	 * @param byteorder The byte order. This is either BIG_ENDIAN or LITTLE_ENDIAN.
	 */
	public EndianOutputStream(OutputStream output, int byteorder) {
		// Super the output stream.
		super(new DataOutputStream(output));
		
		// Since we don't want to cast the entire freaking time, we cache the uncast
		// OutputStream.
		cache = (DataOutputStream) out;
		
		// Set the byteorder.
		setByteorder(byteorder);
	}
	
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param value The next boolean to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeBoolean(boolean)
	 */
	public void writeBoolean(boolean value) throws IOException {
		// Delegate.
		cache.writeBoolean(value);
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param value The next byte to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeByte(int)
	 */
	public void writeByte(int value) throws IOException {
		// Delegate.
		cache.writeByte(value);
	}
	
	/**
	 * Writes a byte to the underlying stream. It's not <code>DataOutput</code> stuff,
	 * but it provides symmetry with <code>EndianInputStream</code>.
	 *
	 * @param value The next real byte to be written to this stream.
	 * @throws IOException When IO errors happen.
	 */
	public void writeByte(byte value) throws IOException {
		// Overload...
		writeByte((int) value);
	}
	
	/**
	 * Writes an unsigned byte to the underlying stream. It's not <code>DataOutput</code>
	 * stuff, but it provides symmetry with <code>EndianInputStream</code>.
	 *
	 * @param value The next unsigned byte to be written to this stream.
	 * @throws IOException When IO errors happen.
	 */
	public void writeUnsignedByte(int value) throws IOException {
		// Overload.
		writeByte(value);
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param value The next short to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeShort(int)
	 */
	public void writeShort(int value) throws IOException {
		// Create a new byte array of two bytes long.
		byte[] bytes = new byte[2];
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Set up the short the big endian way.
			bytes[0] = (byte) ((value >> 8) & 0xFF);
			bytes[1] = (byte) ((value     ) & 0xFF);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Set up the short the little endian way.
			bytes[1] = (byte) ((value >> 8) & 0xFF);
			bytes[0] = (byte) ((value     ) & 0xFF);
		}
		else  throw new InternalError("Unknown byte order.");
		
		// Write out the bytes.
		cache.write(bytes);
	}
	
	/**
	 * Writes a short to the underlying stream. It's not <code>DataOutput</code> stuff,
	 * but it provides symmetry with <code>EndianInputStream</code>.
	 *
	 * @param value The next real short to be written to this stream.
	 * @throws IOException When IO errors happen.
	 */
	public void writeShort(short value) throws IOException {
		// Overload.
		writeShort((int) value);
	}
	
	/**
	 * Writes an unsigned short to the underlying stream. It's not
	 * <code>DataOutput</code> stuff, but it provides symmetry with
	 * <code>EndianInputStream</code>.
	 *
	 * @param value The next unsigned short to be written to this stream.
	 * @throws IOException When IO errors happen.
	 */
	public void writeUnsignedShort(int value) throws IOException {
		// Overload.
		writeShort(value);
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>. This doesn't use byte
	 * ordering, cause no one writes out his Unicode like this :P 
	 *
	 * @param value The next char to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeChar(int)
	 */
	public void writeChar(int value) throws IOException {
		// Delegate.
	    cache.writeChar(value);
	}
	
	/**
	 * Symmerty method for <code>EndianInputStream</code>. This doesn't use byte
	 * ordering, cause no one writes out his Unicode like this :P 
	 *
	 * @param value The next char to be written to this stream.
	 * @throws IOException When IO errors happen.
	 */
	public void writeChar(char value) throws IOException {
		// Overload.
	    cache.writeChar((int) value);
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param value The next int to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeInt(int)
	 */
	public void writeInt(int value) throws IOException {
		// Create a new byte array of four bytes long.
		byte[] bytes = new byte[4];
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Set up the int the big endian way.
			bytes[0] = (byte) ((value >> 24) & 0xFF);
			bytes[1] = (byte) ((value >> 16) & 0xFF);
			bytes[2] = (byte) ((value >>  8) & 0xFF);
			bytes[3] = (byte) ((value      ) & 0xFF);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Set up the int the little endian way.
			bytes[3] = (byte) ((value >> 24) & 0xFF);
			bytes[2] = (byte) ((value >> 16) & 0xFF);
			bytes[1] = (byte) ((value >>  8) & 0xFF);
			bytes[0] = (byte) ((value      ) & 0xFF);
		}
		else  throw new InternalError("Unknown byte order.");
		
		// Write out the bytes.
		cache.write(bytes);
	}
	
	/**
	 * Writes an unsigned int to the underlying stream. It's not <code>DataOutput</code>
	 * stuff, but it provides symmetry with <code>EndianInputStream</code>.
	 *
	 * @param value The next unsigned byte to be written to this stream.
	 * @throws IOException When IO errors happen.
	 */
	public void writeUnsignedInt(long value) throws IOException {
		// Overload.
		writeInt((int) value);
	}
	
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param value The next long to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeLong(long)
	 */
	public void writeLong(long value) throws IOException {
		// Create a new byte array of eight bytes long!!
		byte[] bytes = new byte[8];
		
		// Check which mode we are in.
		if(getByteorder() == BIG_ENDIAN) {
			// Set up the long the big endian way.
			bytes[0] = (byte) ((value >> 56) & 0xFF);
			bytes[1] = (byte) ((value >> 48) & 0xFF);
			bytes[2] = (byte) ((value >> 40) & 0xFF);
			bytes[3] = (byte) ((value >> 32) & 0xFF);
			bytes[4] = (byte) ((value >> 24) & 0xFF);
			bytes[5] = (byte) ((value >> 16) & 0xFF);
			bytes[6] = (byte) ((value >>  8) & 0xFF);
			bytes[7] = (byte) ((value      ) & 0xFF);
		}
		else if(getByteorder() == LITTLE_ENDIAN) {
			// Set up the long the little endian way.
			bytes[7] = (byte) ((value >> 56) & 0xFF);
			bytes[6] = (byte) ((value >> 48) & 0xFF);
			bytes[5] = (byte) ((value >> 40) & 0xFF);
			bytes[4] = (byte) ((value >> 32) & 0xFF);
			bytes[3] = (byte) ((value >> 24) & 0xFF);
			bytes[2] = (byte) ((value >> 16) & 0xFF);
			bytes[1] = (byte) ((value >>  8) & 0xFF);
			bytes[0] = (byte) ((value      ) & 0xFF);
		}
		else  throw new InternalError("Unknown byte order.");
		
		// Write out the bytes.
		cache.write(bytes);
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param value The next float to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeFloat(float)
	 */
	public void writeFloat(float value) throws IOException {
		// Convert the float to an int an write that.
		writeInt(Float.floatToIntBits(value));
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param value The next double to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeDouble(double)
	 */
	public void writeDouble(double value) throws IOException {
		// Convert the double to a long an write that.
		writeLong(Double.doubleToLongBits(value));
	}
	
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param string The next string to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeBytes(String)
	 */
	public void writeBytes(String string) throws IOException {
		// Delegate.
		cache.writeBytes(string);
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param string The next string to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeChars(String)
	 */
	public void writeChars(String string) throws IOException {
		// Delegate.
	    cache.writeChars(string);
	}
	
	/**
	 * See the general contract of <code>DataOutput</code>.
	 *
	 * @param string The next string to be written to this stream.
	 * @throws IOException When IO errors happen.
	 * @see java.io.DataOutput#writeUTF(String)
	 */
	public void writeUTF(String string) throws IOException {
		// Delegate.
	    cache.writeUTF(string);
	}
	
	
	/**
	 * Returns the byte order.
	 * 
	 * @return The byte order used by this stream.
	 */
	public int getByteorder() {
		return byteorder;
	}
	
	/**
	 * Sets the byte order.
	 * 
	 * @param byteorder The byte order used by this stream. 
	 */
	protected void setByteorder(int byteorder) {
		// Check.
		if(byteorder == BIG_ENDIAN || byteorder == LITTLE_ENDIAN)  this.byteorder = byteorder;
		else  throw new IllegalArgumentException("Unsupported byte order.");
	}
	
	
	/**
	 * The <code>OutputStream</code> that <code>FilterOutputStream</code> parent class
	 * delegates to, only uncast.
	 */
	protected DataOutputStream cache;
	
	/**
	 * The byte order used by this stream.
	 */
	protected int byteorder;
	
	
	/**
	 * The Big Endian byte order.
	 */
	public static final int BIG_ENDIAN = 0;
	
	/**
	 * The Little Endian byte order.
	 */
	public static final int LITTLE_ENDIAN = 1;
}