/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// A class that works with streams qualifies as an IO class :P
package com.kirkwarez.kirkpack.io;

// We'll need the IO classes :P
import java.io.*;

/**
 * <p>The <code>Streamer</code> is avery simple class that reads from one stream, and
 * writes to another, the size of the buffer at a time, until the providing stream runs
 * out of data, or the optionally specified length parameter has been reached.</p>
 * 
 * <p>Due to an annoying coincidence, the stream method that streams for a length using
 * the default buffersize does not exist. You can circumvent this by using the stream
 * method that does length and supplying the <code>DEFAULT_BUFFER_SIZE</code> as
 * buffersize. Sorry :^P</p>
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 */
public class Streamer {
	/**
	 * Constructs new <code>Streamer</code> that will read from the given input and
	 * write it to the given output.
	 * 
	 * @param input The stream to read from. May not be null.
	 * @param output The stream to write to. May not be null.
	 */
	public Streamer(InputStream input, OutputStream output) {
		// Set the streams.
		setInputStream(input);
		setOutputStream(output);
	}
	
	
	/**
	 * Streams one stream to another using the given buffer size for the given amount of
	 * bytes.
	 * 
	 * @param buffersize The size of the buffer to use.
	 * @param length The amount of bytes to stream over.
	 * @throws IOException When the streams mess up, or when there is not enough data.
	 */
	public void stream(int buffersize, int length) throws IOException {
		// Start streaming!
		for(int transferred = 0; transferred < length; transferred += buffersize) {
			// How much there's left.
			int left = length - transferred;
			
			// Check whether we're near the end, and set the buffer to the rest if so.
			if(left < buffersize)  buffersize = left;
			
			// Read in the buffer...
			byte[] buffer = read(buffersize);
			
			// Check for buffer underruns (haha).
			if(buffer.length < buffersize)  throw new IOException("Cannot continue streaming: End of stream reached.");
			
			// ...write out the buffer!
			write(buffer);
		}
	}
	
	/**
	 * Streams one stream to another using the given buffer size. Goes on until the input
	 * stream has drained.
	 * 
	 * @param buffersize The size of the buffer to use.
	 * @throws IOException When the streams mess up.
	 */
	public void stream(int buffersize) throws IOException {
		// Boolean for checking done-ness.
		boolean done = false;
		
		// Start streaming!
		while(!done) {
			// Read in the buffer...
			byte[] buffer = read(buffersize);
			
			// ...write out the buffer!
			// We check if there actually IS a buffer to write. If not, the loop ends.
			if(buffer.length > 0)  write(buffer);
			else  done = true;
		}
	}
	
	/**
	 * Streams one stream to another using the default buffer size. Goes on until the
	 * input stream has drained.
	 * 
	 * @throws IOException When the streams mess up.
	 */
	public void stream() throws IOException {
		// Overload to the complete method using the default buffer size.
		stream(DEFAULT_BUFFER_SIZE);
	}
	
	
	/**
	 * Fully reads in the buffer from the input stream, unless the end of the stream is
	 * reached. In that case, a smaller buffer with the remains is returned.
	 * 
	 * @param buffersize the amount to read in.
	 * @return The buffer with the data.
	 * @throws IOException When there is trouble reading.
	 */
	protected byte[] read(int buffersize) throws IOException {
		// Make buffer!
		byte[] buffer = new byte[buffersize];
		
		// Initialise...
		int left = buffersize;
		
		// ...and read!
		while(left > 0) {
			// Read and remember how much we read.
			int read = input.read(buffer, buffersize - left, left);
			
			// Check if we actually read something.
			if(read > 0)  left = left - read;
			else {
				// We're at the end of the file. We'll need to reallocate the buffer
				// with the shorter length...
				byte[] rest = new byte[buffersize - left];
				
				// ...and copy over what's left into this new array.
				System.arraycopy(buffer, 0, rest, 0, buffersize - left);
				
				// Return it, we're done.
				return rest;
			}
		}
		
		// We're done.
		return buffer;
	}
	
	/**
	 * Fully writes out the given buffer to the output stream.
	 * 
	 * @param buffer The buffer to write to the output stream.
	 * @throws IOException When there is trouble writing.
	 */
	protected void write(byte[] buffer) throws IOException {
		// Write out :P
		output.write(buffer);
	}
	
	
	/**
	 * Returns the input stream.
	 * 
	 * @return The stream which will be read from.
	 */
	public InputStream getInputStream() {
		return input;
	}
	
	/**
	 * Sets the input stream.
	 * 
	 * @param input The stream which will be read from.
	 */
	protected void setInputStream(InputStream input) {
		// Check for nullity.
		if(input != null)  this.input = input;
		else throw new NullPointerException("Cannot set input stream: null reference given.");
	}
	
	
	/**
	 * Returns the output stream.
	 * 
	 * @return The stream which will be wriiten to.
	 */
	public OutputStream getOutputStream() {
		return output;
	}
	
	/**
	 * Sets the output stream.
	 * 
	 * @param output The stream which will be written to.
	 */
	protected void setOutputStream(OutputStream output) {
		// Check for nullity.
		if(output != null)  this.output = output;
		else throw new NullPointerException("Cannot set output stream: null reference given.");
	}
	
	
	/**
	 * The stream to read from.
	 */
	protected InputStream input;
	
	/**
	 * The stream to write to.
	 */
	protected OutputStream output;
	
	
	/**
	 * The default buffer size.
	 */
	public static final int DEFAULT_BUFFER_SIZE = 8192;
}