/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// The Myst V: End of Ages encryptor!
package com.kirkwarez.kirkpack.io.eoa.wdysv;

// Import IO.
import java.io.*;

// Import Crypto stuff.
import javax.crypto.*;
import javax.crypto.spec.*;

// Import my IO.
import com.kirkwarez.kirkpack.io.*;

/**
 * This stream streams in MYST V encrypted data and streams out the decrypted data. Since
 * the WDYSV header has the length tag set, one can retrieve the eventual length of the
 * data at any time.
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 */
public class WDYSVInputStream extends FilterInputStream {
	/**
	 * Creates a stream over the given stream.
	 * 
	 * @param input The stream to read from.
	 * @throws IOException When the given stream does not produce EAO data.
	 */
	public WDYSVInputStream(InputStream input) throws IOException {
		// Initialise the superstream.
		super(new EndianInputStream(input, EndianInputStream.LITTLE_ENDIAN));
		
		// Retrieve the uncast version.
		EndianInputStream reader = (EndianInputStream) in;
		
		// Read an int from the inputstream and check if it's the MYST V header.
		if(reader.readInt() == WDYSV_HEADER) {
			// Read and set the length.
			setLength(reader.readInt());
			
			// Catch the Cipher errors.
			try {
				// Construct the cipher.
				SecretKeySpec keyspec = new SecretKeySpec(KEY, "AES");
				Cipher aes = Cipher.getInstance("AES/ECB/NoPadding");
				aes.init(Cipher.DECRYPT_MODE, keyspec);
				
				// Make the encryption reader and make it the cool reader.
				in = new CipherInputStream(in, aes);
			}
			catch(Exception e) {
				// Oh noes a cipher error.
				throw new InternalError("The Cipher wasn't able to initialise.");
			}
		}
		else  throw new IOException("Cannot decrypt EOA data: Wrong header, this file is not a EAO file.");
	}
	
	
	/**
	 * Reads an unsigned byte and returns it as an int. When the length indicated by the
	 * file is reached, no more bytes are read in.
	 * 
	 * @return The byte as an int.
	 * @throws IOException When boo boo goes doo doo. 
	 * @see java.io.InputStream#read()
	 */
	public int read() throws IOException {
		// Check if still need to read.
		if(total < getLength()) {
			// Read.
			int read = super.read();
			
			// Check.
			if(read > 0) {
				// Increase.
				total += 1;
				
				// Return.
				return read;
			}
			else {
				// End of stream? That's not how it should be...
				throw new IOException("End of stream reached before data length was reached: Incomplete EOA file.");
			}
		}
		else {
			// The total is equal (or larger...?) than the header-set length. This means
			// end of stream for us!
			return -1;
		}
	}
	
	/**
	 * Fills the array with bytes from the stream. Not necessarily completely though, but
	 * with at least one. When the length indicated by the file is reached, no more bytes
	 * are read in.
	 * 
	 * @param buffer The buffer to read the data into.
	 * @param offset The starting offset.
	 * @param length The amount of bytes to read.
	 * @return The amount of bytes actually read.
	 * @throws IOException When boo boo goes doo doo. 
	 * @see java.io.InputStream#read(byte[], int, int)
	 */
	public int read(byte buffer[], int offset, int length) throws IOException {
		// Check whether anything should even be read.
		if(length > 0) {
			// Check if we're there yet.
			if(total < getLength()) {
				// Check whether we're about to overflow the length of the unencrypted
				// data.
				if(total + length > getLength()) {
					// Adjust the amount of bytes to read to the number of bytes that are
					// left.
					length = getLength() - total;
				}
				
				// Read.
				int read = super.read(buffer, offset, length);
				
				// Check if there was stuff read.
				if(read > 0) {
					// Increase the total.
					total += read;
					
					// Return.
					return read;
				}
				else {
					// We read -1 because end of stream has been detected. This is not
					// right for this kind of file.
					throw new IOException("End of stream reached before data length was reached: Incomplete EOA file.");
				}
			}
			else {
				// We're at the unencrypted data length. Fake the end of the stream.
				return -1;
			}
		}
		else {
			// Nothing we had to read.
			return 0;
		}
	}
	
	/**
	 * Skips the given amount of bytes, or at least tries to.
	 * 
	 * @return The actual amount of bytes skipped.
	 * @see java.io.InputStream#read(byte[], int, int)
	 * @see java.io.InputStream#skip(long)
	 */
	public long skip(long amount) throws IOException {
		// Check if we're about to overflow the length of the unencrypted data.
		if(total + amount > getLength()) {
			// Set the amount of bytes to skip to the number of bytes that are left.
			amount = getLength() - total;
		}
		
		// Skip.
		long skipped = in.skip(amount);
		
		// Check.
		total += (int) skipped;
		
		// Return.
		return skipped;
	}
	
	
	/**
	 * Returns the number of bytes that can be read without blockage. This version keeps
	 * in mind that there is a preset length in the header.
	 * 
	 * @return The number of bytes that can be read without blockage.
	 * @see java.io.InputStream#available()
	 */
	public int available() throws IOException {
		// Ask the cihper stream. It's available method depends on what it still has
		// left from it's currently decrypted block.
		int available = in.available();
		
		// However, we only want the user to know there is stuff left as long as the
		// length hasn't been reached.
		if(total + available > getLength()) {
			// Set the amount of bytes to skip to the number of bytes that are left.
			available = getLength() - total;
		}
		
		// Return.
		return available;
	}
	
	
	/**
	 * Returns the length of the unencrypted data as indicated by the header of the file.
	 * 
	 * @return The length of the unencrypted data.
	 */
	public int getLength() {
		return length;
	}
	
	/**
	 * Sets the length of the unencrypted data from the indication by the header of the
	 * file.
	 * @param length The length of the unencrypted data.
	 */
	protected void setLength(int length) {
		// Check and set.
		if(length >= 0)  this.length = length;
		else  throw new IllegalArgumentException("Negative length not allowed");
	}
	
	
	/**
	 * The length read from the header.
	 */
	protected int length;
	
	/**
	 * The total read from the stream.
	 */
	protected int total = 0;
	
	
	/**
	 * The key. Ssh!
	 */
	private static byte[] KEY = new byte[]{
			-16, 77, 37, 51, -84, 93, 39, 90, -98, 24, 120, 62, 101, 44, 72, 8
		};
	
	/**
	 * The WDYSV header.
	 */
	public static final int WDYSV_HEADER = 0x0D874288;
}