/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// The Myst V: End of Ages encryptor!
package com.kirkwarez.kirkpack.io.eoa.wdysv;

// Import IO. 
import java.io.*;

//Import Crypto stuff.
import javax.crypto.*;
import javax.crypto.spec.*;

// Import my IO.
import com.kirkwarez.kirkpack.io.*;

/**
 * This stream streams in plain input data and streams out MYST V encrypted data. Note
 * that a length attribute is required upon construction. This is because the WDYSV
 * header requires the length of the unencrypted data to be set, and it cannot be
 * determined from the stream. The encryption requires blocks of 16 bytes to work, and
 * any padding will be done with zeroes.
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 */
public class WDYSVOutputStream extends FilterOutputStream {
	/**
	 * Creates a stream over the given stream.
	 * 
	 * @param output The stream to write to.
	 * @param length The length of the data that will be written into this stream
	 * (needed for the header, and also for padding).
	 * @throws IOException When the given stream does not cooperate (we need to write the
	 * header and the length attribute).
	 */
	public WDYSVOutputStream(OutputStream output, int length) throws IOException {
		// Initialise the superstream.
		super(new EndianOutputStream(output, EndianOutputStream.LITTLE_ENDIAN));
		
		// Retrieve the uncast version.
		EndianOutputStream writer = (EndianOutputStream) out;
		
		// Write the MYST V header to the output stream.
		writer.writeInt(WDYSV_HEADER);
		
		// Write and set the length.
		writer.writeInt(length);
		setLength(length);
		
		// Catch the Cipher errors.
		try {
			// Construct the cipher.
			SecretKeySpec keyspec = new SecretKeySpec(KEY, "AES");
			Cipher aes = Cipher.getInstance("AES/ECB/NoPadding");
			aes.init(Cipher.ENCRYPT_MODE, keyspec);
			
			// Make the encryption writer and make it the cool writer.
			out = new CipherOutputStream(out, aes);
		}
		catch(Exception e) {
			// Oh noes a cipher error.
			throw new InternalError("The Cipher wasn't able to initialise.");
		}
	}
	
	
	/**
	 * Writes the given byte (well, int) to the underlying steam. When the length
	 * indicated at construction time is reached, no more bytes are written out.
	 * 
	 * @param value An int that represents a byte to be written out.
	 * @see java.io.OutputStream#write(int)
	 * @throws IOException When, you know, IO stuff.
	 */
	public void write(int value) throws IOException {
		// Check whether we're at the length yet.
		if(total < getLength()) {
			// Write.
			super.write(value);
			
			// Increase total.
			total++;
			
			// If the total is now equal to the length we'll need to write out the
			// padding as well.
			if(total == length)  writePadding();
		}
		else  throw new IOException("Cannot write data beyond the given data length.");
	}
	
	/**
	 * Writes the given byte array to the underlying steam using the given subselection.
	 * When the length indicated at construction time is reached, no more bytes are
	 * written out.
	 * 
	 * @param buffer The byte buffer with data.
	 * @param offset The starting point of where to start getting the data.
	 * @param length The amount of data to write, taken from offset on.
	 * @see java.io.OutputStream#write(byte[], int, int)
	 */
	public void write(byte buffer[], int offset, int length) throws IOException {
		// Check whether anything should be written.
		if(length > 0) {
			// Check whether we're at the length yet.
			if(total < getLength()) {
				// Check whether we're about to overflow the length of the unencrypted
				// data.
				if(total + length > getLength()) {
					// Adjust the amount of bytes to write to the number of bytes that
					// are left.
					length = getLength() - total;
				}
				
				// Write.
				// Don't super, because that silly stream writes each byte on its own.
				out.write(buffer, offset, length);
				
				// Increase total.
				total += length;
				
				// Check whether we're there yet NOW, and if so pad until the block
				// is filled.
				if(total == length)  writePadding();
			}
			else {
				// We're at the end, we shouldn't write anymore.
				throw new IOException("Cannot write data beyond the given data length.");
			}
		}
	}
	
	
	/**
	 * Writes out padding to fill up the blocksize. The blocksize is 16 bytes.
	 * 
	 * @throws IOException When writing fails and stuff.
	 */
	protected void writePadding() throws IOException {
		// Calculate the end bit and write it out.
		out.write(new byte[BLOCKSIZE - getLength() % BLOCKSIZE]);
	}
	
	
	/**
	 * Returns the length of the unencrypted data, as written to the fileheader.
	 * 
	 * @return length The length of the unencrypted data.
	 */
	public int getLength() {
		return length;
	}
	
	/**
	 * Sets the length of the unencrypted data, as written to the fileheader.
	 * 
	 * @param length The length of the unencrypted data.
	 */
	protected void setLength(int length) {
		this.length = length;
	}
	
	
	/**
	 * The length of the unencrypted data.
	 */
	protected int length;
	
	/**
	 * The amount of bytes written out.
	 */
	protected int total = 0;
	
	
	/**
	 * The key. Ssh!
	 */
	private static byte[] KEY = new byte[]{
			-16, 77, 37, 51, -84, 93, 39, 90, -98, 24, 120, 62, 101, 44, 72, 8
		};
	
	/**
	 * The blocksize.
	 */
	protected static final int BLOCKSIZE = 16;
	
	/**
	 * The WDYSV header.
	 */
	public static final int WDYSV_HEADER = 0x0D874288;
}