/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// This class filters files, thus IO.
package com.kirkwarez.kirkpack.io.filter.file;

// Import the IO classes.
import java.io.*;

/**
 * This class is used for filtering out files with the given extension(s) when using
 * <code>java.io.File.listFiles(java.io.FileFilter)</code> or other things requiring
 * file filters.
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 * @see java.io.File
 */
public class ExtensionFileFilter extends NormalFileFilter implements FileFilter {
	/**
	 * Constructs a new filter that filters out files with the given extension.
	 * 
	 * @param extension The extension, without a period.
	 */
	public ExtensionFileFilter(String extension) {
		// Create a new mini array.
		extensions = new String[1];
		
		// Put it in there.
		extensions[0] = extension;
	}
	
	/**
	 * Constructs a new filter that filters out files with the given extensions.
	 * 
	 * @param extensions The extensions, without periods.
	 */
	public ExtensionFileFilter(String[] extensions) {
		// Set the extensions to the given extensions.
		this.extensions = extensions;
	}
	
	
	/**
	 * Returns whether when the passed file should be included in the listing. If it
	 * is a file with any of the given extensions, it should.
	 * 
	 * @param file The file to be checked upon.
	 * @return Whether the file is an accepted file.
	 */
	public boolean accept(File file) {
		// Return whether this file is a normal file.
		if(super.accept(file)) {
			// Check for any of the extensions.
			for(int index = 0; index < extensions.length; index++) {
				// Check for the current extension.
				if(file.getPath().endsWith("." + extensions[index]))  return true;
			}
			
			// If we get here there was no match.
			return false;
		}
		else {
			// It's not even a normal file!
			return false;
		}
	}
	
	
	/**
	 * The extension to filter on.
	 */
	protected String[] extensions;
}