/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// The Uru encryptor!
package com.kirkwarez.kirkpack.io.uru.wdys;

// Import IO and utility classes! 
import java.io.*;
import java.util.*;

// Import my IO.
import com.kirkwarez.kirkpack.io.*;

/**
 * This stream streams in Uru WDYS encrypted data and streams out the decrypted data.
 * Since the WDYS header has the length tag set, one can retrieve the eventual length of
 * the data at any time.
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 */
public class WDYSInputStream extends FilterInputStream {
	/**
	 * Creates a stream over the given stream.
	 * 
	 * @param input The stream to read from.
	 * @throws IOException When the given stream does not produce Uru data.
	 */
	public WDYSInputStream(InputStream input) throws IOException {
		// Initialise the superstream.
		super(new EndianInputStream(input, EndianInputStream.LITTLE_ENDIAN));
		
		// Retrieve the uncast version.
		EndianInputStream reader = (EndianInputStream) in;
		
		// Read 12 bytes from the inputstream.
		byte[] header = reader.readFully(12);
		
		// Check if it's the Uru header.
		if(Arrays.equals(header, WDYS_HEADER.getBytes("US-ASCII"))) {
			// Read and set the length.
			setLength(reader.readInt());
		}
		else  throw new IOException("Cannot decrypt Uru data: Wrong header, this file is not an Uru file.");
	}
	
	
	/**
	 * Reads an unsigned byte and returns it as an int. When the length indicated by the
	 * file is reached, no more bytes are read in.
	 * 
	 * @return The byte as an int.
	 * @throws IOException When boo boo goes doo doo. 
	 * @see java.io.InputStream#read()
	 */
	public int read() throws IOException {
		// Check if we still need to read.
		if(total < getLength()) {
			// The buffer we're going to use.
			byte[] buffer = new byte[1];
			
			// "Read".
			int read = readFromBuffer(buffer, 0, 1);
			
			// Check.
			if(read > 0) {
				// Increase.
				total += 1;
				
				// Return.
				return buffer[0];
			}
			else {
				// End of stream? That's not how it should be...
				throw new IOException("End of stream reached before data length was reached: Incomplete Uru file.");
			}
		}
		else {
			// The total is equal (or larger...?) than the header-set length. This means
			// end of stream for us!
			return -1;
		}
	}
	
	/**
	 * Fills the array with bytes from the stream. Not necessarily completely though, but
	 * with at least one. When the length indicated by the file is reached, no more bytes
	 * are read in.
	 * 
	 * @param buffer The buffer to read the data into.
	 * @param offset The starting offset.
	 * @param length The amount of bytes to read.
	 * @return The amount of bytes actually read.
	 * @throws IOException When boo boo goes doo doo. 
	 * @see java.io.InputStream#read(byte[], int, int)
	 */
	public int read(byte buffer[], int offset, int length) throws IOException {
		// Check whether anything should even be read.
		if(length > 0) {
			// Check if we're there yet.
			if(total < getLength()) {
				// Check whether we're about to overflow the length of the unencrypted
				// data.
				if(total + length > getLength()) {
					// Adjust the amount of bytes to read to the number of bytes that are
					// left.
					length = getLength() - total;
				}
				
				// "Read".
				int read = readFromBuffer(buffer, offset, length);
				
				// Check if there was stuff read.
				if(read > 0) {
					// Increase the total.
					total += read;
					
					// Return.
					return read;
				}
				else {
					// We read -1 because end of stream has been detected. This is not
					// right for this kind of file.
					throw new IOException("End of stream reached before data length was reached: Incomplete Uru file.");
				}
			}
			else {
				// We're at the unencrypted data length. Fake the end of the stream.
				return -1;
			}
		}
		else {
			// Nothing we had to read.
			return 0;
		}
	}
	
	/**
	 * Skips the given amount of bytes, or at least tries to.
	 * 
	 * @return The actual amount of bytes skipped.
	 * @see java.io.InputStream#read(byte[], int, int)
	 * @see java.io.InputStream#skip(long)
	 */
	public long skip(long amount) throws IOException {
		// Due to the length header being an int we can skip so much bytes.
		int skip = (int) amount;
		
		// Check whether we should skip more than just our buffers.
		if(skip > unread) {
			// Reset the length to what we've got.
			skip = unread;
		}
		
		// Check if we're about to overflow the length of the unencrypted data.
		if(total + skip > getLength()) {
			// Set the amount of bytes to skip to the number of bytes that are left.
			skip = getLength() - total;
		}
		
		// Skip bytes by changing the amount of unread ones.
		unread -= skip;
		
		// Check.
		total += skip;
		
		// Return.
		return skip;
	}
	
	
	/**
	 * Returns the number of bytes that can be read without blockage. This version keeps
	 * in mind that there is a preset length in the header, and that there is at most
	 * 8 bytes that can be read without blocking (decrypting block size).
	 * 
	 * @return The number of bytes that can be read without blockage.
	 * @see java.io.InputStream#available()
	 */
	public int available() throws IOException {
		// The initial amount of bytes left to be read without blockage. This is the
		// amount of unread bytes from the decrypted buffer.
		int available = unread;
		
		// We only want the user to know there is stuff left as long as the length hasn't
		// been reached.
		if(total + available > getLength()) {
			// Set the amount of bytes to skip to the number of bytes that are left.
			available = getLength() - total;
		}
		
		// Return.
		return available;
	}
	
	
	/**
	 * Reads length bytes from the buffers and puts them in the given array from offset
	 * on.
	 * 
	 * @param buffer The buffer to read the data into.
	 * @param offset The starting offset.
	 * @param length The amount of bytes to read.
	 * @return The amount of bytes actually read. This method will only read until the
	 * buffers have been depleted.
	 * @throws IOException When your HDD explodes.
	 */
	protected synchronized int readFromBuffer(byte buffer[], int offset, int length) throws IOException {
		// Check whether the buffers are empty.
		if(unread == 0) {
			// Retrieve the uncast version of the stream.
			EndianInputStream reader = (EndianInputStream) in;
			
			// The blocks we'll be reading in.
			byte[][] blocks = new byte[2][];
			
			// Try to read.
			try {
				// Fill the buffers.
				blocks[0] = reader.readFully(4);
				blocks[1] = reader.readFully(4);
			}
			catch(EOFException e) {
				// Aw crap. This file is bad file. We return -1 for end of stream, the
				// calling method translates this to "bad" and throws the right
				// exception. It doesn't matter the stream is left in an inconsistant
				// state. Every consecutive call will result in the exception.
				return -1;
			}
			
			// Decrypt the buffers.
			blocks = decrypt(blocks);
			
			// Copy the blocks into our one-size array.
			System.arraycopy(blocks[0], 0, block, 0, 4);
			System.arraycopy(blocks[1], 0, block, 4, 4);
			
			// Reset the unread setting.
			unread = 8;
		}
		
		// Check whether we should read more than just from our buffers.
		if(length > unread) {
			// Reset the length to what we've got.
			length = unread;
		}
		
		// Copy length bytes from the block array into the given buffer array.
		System.arraycopy(block, 8 - unread, buffer, offset, length);
		
		// Decrease unread.
		unread -= length;
		
		// Return how much we've read.
		return length;
	}
	
	/**
	 * The xTEA decryption algorithm.
	 * 
	 * @param v The encrypted input blocks. 
	 * @return The decrypted input blocks. These are not the same byte arrays, however
	 * the container byte array is.
	 */
	protected byte[][] decrypt(byte[][] v) {
		// Set the helper values for the algorithm.
		int sum = 0xC6EF3720;
		int delta = 0x9E3779B9;
		
		// Initialise the values.
		int v0 = byteArrayToInt(v[0]);
		int v1 = byteArrayToInt(v[1]);
		
		// Repeat for 32 steps (the standard).
		for(int count = 0; count < 32; count++) {
			// Exercise algorithm.
			v1 -= ((v0 << 4 ^ v0 >>> 5) + v0) ^ (sum + KEY[sum >>> 11 & 3]);
			sum -= delta;
			v0 -= ((v1 << 4 ^ v1 >>> 5) + v1) ^ (sum + KEY[sum & 3]);
		}
		
		// Convert the final values into usuable stuff for this stream.
		v[0] = intToByteArray(v0);
		v[1] = intToByteArray(v1);
		
		// Return the final values.
		return v;
	}
	
	
	/**
	 * Makes an int out of a byte array. It'll be using Little Endian, for the sake of
	 * the decryption algorithm implemented by this class.
	 * 
	 * @param value The int to turn into a byte array.
	 * @return The byte array that used to be the int.
	 */
	protected byte[] intToByteArray(int value) {
		// Make the byte array from the int and return it.
		return new byte[]{
			(byte) ((value      ) & 0xFF),
			(byte) ((value >>  8) & 0xFF),
			(byte) ((value >> 16) & 0xFF),
			(byte) ((value >> 24) & 0xFF)
		};
	}
	
	/**
	 * Makes a byte array of an int. It'll be using Little Endian, for the sake of
	 * the decryption algorithm implemented by this class.
	 * 
	 * @param bytes The byte array to turn into an int.
	 * @return The int that used to be a byte array.
	 */
	protected int byteArrayToInt(byte[] bytes) {
		// Make the int from the byte array and return it.
		return
			(
				((bytes[0] & 0xFF)      ) |
				((bytes[1] & 0xFF) <<  8) |
				((bytes[2] & 0xFF) << 16) |
				((bytes[3] & 0xFF) << 24)
				
			);
	}
	
	
	/**
	 * Returns the length of the unencrypted data as indicated by the header of the file.
	 * 
	 * @return The length of the unencrypted data.
	 */
	public int getLength() {
		return length;
	}
	
	/**
	 * Sets the length of the unencrypted data from the indication by the header of the
	 * file.
	 * @param length The length of the unencrypted data.
	 */
	protected void setLength(int length) {
		// Check and set.
		if(length >= 0)  this.length = length;
		else  throw new IllegalArgumentException("Negative length not allowed");
	}
	
	
	/**
	 * The xTEA dual block buffer.
	 */
	protected byte[] block = new byte[8];
	
	/**
	 * How much there's left to be read from the buffers. A value between 0 and 8.
	 */
	protected int unread = 0;
	
	
	/**
	 * The length read from the header.
	 */
	protected int length;
	
	/**
	 * The total read from the stream.
	 */
	protected int total = 0;
	
	
	/**
	 * The key. Ssh!
	 */
	private static int[] KEY = new int[]{
			0x6c0a5452,
			0x03827d0f,
			0x3a170b92,
			0x16db7fc2
		};
	
	/**
	 * The WDYS header.
	 */
	public static final String WDYS_HEADER = "whatdoyousee";
}