/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// The Uru encryptor!
package com.kirkwarez.kirkpack.io.uru.wdys;

// Import daddah! 
import java.io.*;

// Import my IO.
import com.kirkwarez.kirkpack.io.*;

/**
 * This stream streams in plain input data and streams out Uru WDYS encrypted data. Note
 * that a length attribute is required upon construction. This is because the WDYS
 * header requires the length of the unencrypted data to be set, and it cannot be
 * determined from the stream. The encryption requires blocks of 2 times 4 bytes to work,
 * and any padding will be done with zeroes.
 * 
 * @author Berend "Kirk" Wouda
 * @version 3.00
 * @since 3.00
 */
public class WDYSOutputStream extends FilterOutputStream {
	/**
	 * Creates a stream over the given stream.
	 * 
	 * @param output The stream to write to.
	 * @param length The length of the data that will be written into this stream
	 * (needed for the header, and also for padding).
	 * @throws IOException When the given stream does not cooperate (we need to write the
	 * header and the length attribute).
	 */
	public WDYSOutputStream(OutputStream output, int length) throws IOException {
		// Initialise the superstream.
		super(new EndianOutputStream(output, EndianOutputStream.LITTLE_ENDIAN));
		
		// Retrieve the uncast version.
		EndianOutputStream writer = (EndianOutputStream) out;
		
		// Write the Uru header to the output stream.
		writer.write(WDYS_HEADER.getBytes("US-ASCII"));
		
		// Write and set the length.
		writer.writeInt(length);
		setLength(length);
	}
	
	
	/**
	 * Writes the given byte (well, int) to the underlying steam. When the length
	 * indicated at construction time is reached, no more bytes are written out.
	 * 
	 * @param value An int that represents a byte to be written out.
	 * @see java.io.OutputStream#write(int)
	 * @throws IOException When, you know, IO stuff.
	 */
	public void write(int value) throws IOException {
		// Check whether we're at the length yet.
		if(total < getLength()) {
			// The buffer we're going to use.
			byte[] buffer = new byte[]{
				(byte) value	
			};
			
			// "Write".
			writeToBuffer(buffer, 0, 1);
			
			// Increase total.
			total++;
			
			// If the total is now equal to the length we'll need to write out the
			// padding as well.
			if(total == getLength())  writePadding();
		}
		else  throw new IOException("Cannot write data beyond the given data length.");
	}
	
	/**
	 * Writes the given byte array to the underlying steam using the given subselection.
	 * When the length indicated at construction time is reached, no more bytes are
	 * written out.
	 * 
	 * @param buffer The byte buffer with data.
	 * @param offset The starting point of where to start getting the data.
	 * @param length The amount of data to write, taken from offset on.
	 * @see java.io.OutputStream#write(byte[], int, int)
	 */
	public void write(byte buffer[], int offset, int length) throws IOException {
		// Check whether anything should be written.
		if(length > 0) {
			// Check whether we're at the length yet.
			if(total < getLength()) {
				// Check whether we're about to overflow the length of the unencrypted
				// data.
				if(total + length > getLength()) {
					// Adjust the amount of bytes to write to the number of bytes that
					// are left.
					length = getLength() - total;
				}
				
				// "Write".
				writeToBuffer(buffer, offset, length);
				
				// Increase total.
				total += length;
				
				// Check whether we're there yet NOW, and if so pad until the block
				// is filled.
				if(total == getLength())  writePadding();
			}
			else {
				// We're at the end, we shouldn't write anymore.
				throw new IOException("Cannot write data beyond the given data length.");
			}
		}
	}
	
	
	/**
	 * Writes out padding to fill up the blocksize. The blocksize is 8 bytes.
	 * 
	 * @throws IOException When writing fails and stuff.
	 */
	protected void writePadding() throws IOException {
		// Calculate the end bit...
		int left = 8 - getLength() % 8;
		
		// ...and write it out.
		writeToBuffer(new byte[left], 0, left);
	}
	
	
	/**
	 * Writes length bytes from the given array to the buffers from offset on.
	 * 
	 * @param buffer The buffer to write the data from.
	 * @param offset The starting offset.
	 * @param length The amount of bytes to write.
	 * @throws IOException When your HDD explodes.
	 */
	protected synchronized void writeToBuffer(byte buffer[], int offset, int length) throws IOException {
		// How much there is left to write.
		int towrite;
		
		// Cycle through until we've reached length.
		for(int left = length; left > 0; left -= towrite) {
			// Cache how much there's left.
			towrite = left;
			
			// Check whether we should write more than just to our buffers.
			if(towrite > unwritten) {
				// Reset the length to what we've got.
				towrite = unwritten;
			}
			
			// Copy towrite bytes into the block array.
			System.arraycopy(buffer, offset + length - left, block, 8 - unwritten, towrite);
			
			// Decrease unwritten.
			unwritten -= towrite;
			
			// Check whether the buffers are full.
			if(unwritten == 0) {
				// Retrieve the uncast version of the stream.
				EndianOutputStream writer = (EndianOutputStream) out;
				
				// Make the block arrays were's going to write with.
				byte[][] blocks = new byte[2][4];
				
				// Split the block array.
				System.arraycopy(block, 0, blocks[0], 0, 4);
				System.arraycopy(block, 4, blocks[1], 0, 4);
				
				// Encrypt the buffers.
				blocks = encrypt(blocks);
				
				// Write out the buffers.
				writer.write(blocks[0]);
				writer.write(blocks[1]);
				
				// Reset the unwritten setting.
				unwritten = 8;
			}
		}
	}
	
	
	/**
	 * The xTEA encryption algorithm.
	 * 
	 * @param v The encrypted input blocks. 
	 * @return The decrypted input blocks. These are not the same byte arrays, however
	 * the container byte array is.
	 */
	protected byte[][] encrypt(byte[][] v) {
		// Set the helper values for the algorithm.
		int sum = 0;
		int delta = 0x9E3779B9;
		
		// Initialise the values.
		int v0 = byteArrayToInt(v[0]);
		int v1 = byteArrayToInt(v[1]);
		
		// Repeat for 32 steps (the standard).
		for(int count = 0; count < 32; count++) {
			// Exercise algorithm.
			v0 += ((v1 << 4 ^ v1 >>> 5) + v1) ^ (sum + KEY[sum & 3]);
			sum += delta;
			v1 += ((v0 << 4 ^ v0 >>> 5) + v0) ^ (sum + KEY[sum >>> 11 & 3]);
		}
		
		// Convert the final values into usuable stuff for this stream.
		v[0] = intToByteArray(v0);
		v[1] = intToByteArray(v1);
		
		// Return the final values.
		return v;
	}
	
	
	/**
	 * Makes an int out of a byte array. It'll be using Little Endian, for the sake of
	 * the decryption algorithm implemented by this class.
	 * 
	 * @param value The int to turn into a byte array.
	 * @return The byte array that used to be the int.
	 */
	protected byte[] intToByteArray(int value) {
		// Make the byte array from the int and return it.
		return new byte[]{
			(byte) ((value      ) & 0xFF),
			(byte) ((value >>  8) & 0xFF),
			(byte) ((value >> 16) & 0xFF),
			(byte) ((value >> 24) & 0xFF)
		};
	}
	
	/**
	 * Makes a byte array of an int. It'll be using Little Endian, for the sake of
	 * the decryption algorithm implemented by this class.
	 * 
	 * @param bytes The byte array to turn into an int.
	 * @return The int that used to be a byte array.
	 */
	protected int byteArrayToInt(byte[] bytes) {
		// Make the int from the byte array and return it.
		return
			(
				((bytes[0] & 0xFF)      ) |
				((bytes[1] & 0xFF) <<  8) |
				((bytes[2] & 0xFF) << 16) |
				((bytes[3] & 0xFF) << 24)
				
			);
	}
	
	
	/**
	 * Returns the length of the unencrypted data, as written to the fileheader.
	 * 
	 * @return length The length of the unencrypted data.
	 */
	public int getLength() {
		return length;
	}
	
	/**
	 * Sets the length of the unencrypted data, as written to the fileheader.
	 * 
	 * @param length The length of the unencrypted data.
	 */
	protected void setLength(int length) {
		this.length = length;
	}
	
	
	/**
	 * The xTEA dual block buffer.
	 */
	protected byte[] block = new byte[8];
	
	/**
	 * How much there's yet to be written to the buffers to make them full.
	 * A value between 0 and 8.
	 */
	protected int unwritten = 8;
	
	
	/**
	 * The length of the unencrypted data.
	 */
	protected int length;
	
	/**
	 * The amount of bytes written out.
	 */
	protected int total = 0;
	
	
	/**
	 * The key. Ssh!
	 */
	private static int[] KEY = new int[]{
			0x6c0a5452,
			0x03827d0f,
			0x3a170b92,
			0x16db7fc2
		};
	
	/**
	 * The WDYS header.
	 */
	public static final String WDYS_HEADER = "whatdoyousee";
}