/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// Has to do with batching.
package com.kirkwarez.wdysdv.batch;

// Import the IO and utility classes.
import java.io.*;
import java.util.*;

// Import the KirkPack Uru and EAO stuff.
import com.kirkwarez.kirkpack.io.uru.wdys.*;
import com.kirkwarez.kirkpack.io.eoa.wdysv.*;
import com.kirkwarez.kirkpack.io.Streamer;

/**
 * The <code>Processor</code> processes tasks taken from batches.
 * 
 * @author Berend "Kirk" Wouda
 * @version 5.00
 * @since 5.00
 */
public class Processor {
	/**
	 * Constructs a new <code>Processor</code>.
	 */
	public Processor() {
		// Nothing, apparently.
	}
	
	
	/**
	 * Processes the given batch.
	 * 
	 * @param batch The batch to process.
	 * @throws IOException When the files from the tasks have problems.
	 */
	public void process(Batch batch) throws IOException {
		// Get the iterator.
		Iterator iterator = batch.getIterator();
		
		// Go through the tasks in the batch.
		while(iterator.hasNext()) {
			// Process the task.
			process((Task) iterator.next());
			
			// Remove the task safely.
			iterator.remove();
		}
	}
	
	/**
	 * Processes the given task.
	 * 
	 * @param task The task to process.
	 * @throws IOException When the files from the tasks have problems.
	 */
	public void process(Task task) throws IOException {
		// Construct the streams.
		InputStream reader;
		OutputStream writer;
		
		// Check the mode.
		if(task.getMode() == Task.ENCRYPT) {
			// Open up the stream for reading.
			reader = new BufferedInputStream(new FileInputStream(task.getInputFile()));
			
			// Check which stream to open up the stream for writing.
			if(task.getType() == Task.URU)  writer = new WDYSOutputStream(new FileOutputStream(task.getOutputFile()), (int) task.getInputFile().length());
			else if(task.getType() == Task.EOA)  writer = new WDYSVOutputStream(new FileOutputStream(task.getOutputFile()), (int) task.getInputFile().length());
			else  throw new InternalError("Cannot process task: Invalid encryption type.");
		}
		else if(task.getMode() == Task.DECRYPT) {
			// Check which stream to open up the stream for reading.
			if(task.getType() == Task.URU)  reader = new WDYSInputStream(new FileInputStream(task.getInputFile()));
			else if(task.getType() == Task.EOA)  reader = new WDYSVInputStream(new FileInputStream(task.getInputFile()));
			else  throw new InternalError("Cannot process task: Invalid encryption type.");
			
			// Open up the stream for writing.
			writer = new BufferedOutputStream(new FileOutputStream(task.getOutputFile()));
		}
		else  throw new InternalError("Cannot process task: Invalid operation mode.");
		
		// Read and write.
		Streamer streamer = new Streamer(reader, writer);
		streamer.stream();
		
		// Close the streams.
		reader.close();
		writer.close();
	}
}