/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// Task classes are batchable. 
package com.kirkwarez.wdysdv.batch;

// Import the File class. We use it to represent the input and output files.
import java.io.File;

/**
 * <p>The <code>Task</code> class is a class that represents a batchable task. Such an
 * item has an input and output filename, a type (MYST V: EOA or Uru) and a mode (encrypt
 * or * decrypt). It can be fed to a processor that will process this batch item (encrypt
 * or decrypt the input file into the output file using the specified type of
 * encryption).</p>
 * 
 * <p><code>Task</code>s are immutable, because the program doesn't edit them, but
 * removes them and puts the data into the user's hands for reentry.</p>
 * 
 * @author Berend "Kirk" Wouda
 * @version 5.00
 * @since 5.00
 */
public class Task {
	/**
	 * Create a new task.
	 * 
	 * @param input The input <code>File</code>.
	 * @param output The output <code>File</code>.
	 * @param type The game type.
	 * @param mode The encryption mode.
	 */
	public Task(File input, File output, int type, int mode) {
		// Set them.
		setInputFile(input);
		setOutputFile(output);
		setType(type);
		setMode(mode);
	}
	
	
	/**
	 * Returns the input file.
	 * 
	 * @return The input file.
	 */
	public File getInputFile() {
		return input;
	}
	
	/**
	 * Sets the input file to the given <code>File</code>. May not be null.
	 * 
	 * @param input The input file.
	 */
	protected void setInputFile(File input) {
		// Check for nullity.
		if(input != null)  this.input = input;
		else  throw new IllegalArgumentException("Null input file not allowed.");
	}
	
	
	/**
	 * Returns the output file.
	 * 
	 * @return The output file.
	 */
	public File getOutputFile() {
		return output;
	}
	
	/**
	 * Sets the output file to the given <code>File</code>.
	 * 
	 * @param output The input file.
	 */
	protected void setOutputFile(File output) {
		// Check for nullity.
		if(output != null)  this.output = output;
		else  throw new IllegalArgumentException("Null output file not allowed.");
	}
	
	
	/**
	 * Returns the type of encryption, which is either EOA or URU.
	 * 
	 * @return The type of encryption.
	 */
	public int getType() {
		return type;
	}
	
	/**
	 * Sets the type of encryption, to either EOA or URU.
	 * 
	 * @param type The type of encryption.
	 */
	protected void setType(int type) {
		if(type == URU || type == EOA)  this.type = type;
		else  throw new IllegalArgumentException("Cannot set type: Invalid type given.");
	}
	
	
	/**
	 * Returns the mode of operation, which is either ENCRYPT or DECRYPT.
	 * 
	 * @return The mode of operation.
	 */
	public int getMode() {
		return mode;
	}
	
	/**
	 * Sets the mode of operation to either ENCRYPT or DECRYPT.
	 * 
	 * @param mode The mode of operation.
	 */
	protected void setMode(int mode) {
		if(mode == ENCRYPT || mode == DECRYPT)  this.mode = mode;
		else  throw new IllegalArgumentException("Cannot set mode: Invalid mode given.");
	}
	
	
	/**
	 * Returns a string representation of this object, for use in the list.
	 * 
	 * @return What do you think, smartypants? :^P
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		// Get the type and mode.
		String type = (getType() == URU) ? "Uru" : "MYST V";
		String mode = (getMode() == ENCRYPT) ? "Encrypt" : "Decrypt";
		
		// Return something pretty.
		return getInputFile().getName() + " -> " + getOutputFile().getName() + " (Type: " + type + ", Mode: " + mode + ")";
	}
	
	
	/**
	 * The input file.
	 */
	protected File input;
	
	/**
	 * The output file.
	 */
	protected File output;
	
	
	/**
	 * The type of encryption. This is either EOA or URU. End Of Ages uses AES ECB with
	 * an 128 bit key, Uru uses xTEA.
	 */
	protected int type;
	
	
	/**
	 * The mode. This is either ENCRYPT or DECRYPT.
	 */
	protected int mode;
	
	
	/**
	 * The MYST V: End of Ages encryption type.
	 */
	public static final int EOA = 0;
	
	/**
	 * The Uru encryption type.
	 */
	public static final int URU = 1;
	
	
	/**
	 * The encryption mode.
	 */
	public static final int ENCRYPT = 0;
	
	/**
	 * The decryption mode.
	 */
	public static final int DECRYPT = 1;
}