/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// This is a control class.
package com.kirkwarez.wdysdv.control;

// Import the AWT classes.
import java.awt.*;
import java.awt.event.*;

// Import the IO classes.
import java.io.*;

// Import the utility classes. ArrayList HO!
import java.util.*;

// Import the control class.
import com.kirkwarez.wdysdv.gui.DisplayFrame;

// Import my awesome streams.
import com.kirkwarez.kirkpack.io.uru.wdys.*;
import com.kirkwarez.kirkpack.io.eoa.wdysv.*;

// Import Task for its statics.
import com.kirkwarez.wdysdv.batch.Task;

/**
 * This class controls the seperate windows and processes their events.
 * 
 * @author Berend "Kirk" Wouda
 * @version 5.00
 * @since 5.00
 */
public class DisplayControl extends WindowAdapter implements ActionListener {
	/**
	 * Constructs a new <code>DisplayControl</code>.
	 */
	public DisplayControl() {
		// Make a new window list.
		windowlist = new ArrayList();
	}
	
	
	/**
	 * Closes the <code>DisplayWindow</code> indicated by the given
	 * <code>WindowEvent</code> after doing stuff.
	 * 
	 * @param event The event that happened.
	 */
	public void windowClosing(WindowEvent event) {
		// Retrieve the frame.
		DisplayFrame frame = (DisplayFrame) event.getWindow();
		
		// Remove the window from the list.
		windowlist.remove(frame);
		
		// Hide the Window.
		frame.setVisible(false);
		
		// Dump the Window.
		frame.dispose();
	}
	
	
	/**
	 * Processes an event created by an extra window because of user interaction.
	 * 
	 * @param event The event that happened.
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent event) {
		// Retrieve the action command from the event. 
		String command = event.getActionCommand();
		
		// Figure out the window the event happened on.
		Button button = (Button) event.getSource();
		DisplayFrame frame = (DisplayFrame) button.getParent();
		
		// Catch any errors that may occur and want to stop program execution.
		try {
			// Process what the event/user wants us to do.
			if(command.equals("SAVEURUENCRYPTED")) {
				// Let the user browse for a file and save as that file.
				saveEncrypted(Task.URU, frame.getText(), browse(frame));
			}
			else if(command.equals("SAVEEOAENCRYPTED")) {
				// Let the user browse for a file and save as that file.
				saveEncrypted(Task.EOA, frame.getText(), browse(frame));
			}
			else if(command.equals("SAVEDECRYPTED")) {
				// Let the user browse for a file and save as that file.
				saveDecrypted(frame.getText(), browse(frame));
			}
			else {
				// This is bad.
			}
		}
		catch(IOException e) {
			// Alert the user.
			frame.popupError("An error occured while interacting with a file: " + e.getMessage());
		}
		catch(Exception e) {
			// Alert the user.
			frame.popupError("Something totally unexpected happened: " + e.getMessage());
		}
	}
	
	
	/**
	 * Encrypts the given text into the correct type of encryption, and writes it to the
	 * given file.
	 * 
	 * @param type The encryption type to use.
	 * @param text The text to encrypt and save.
	 * @param file The file to save to.
	 * @throws IOException When streamstuffs go wrong.
	 */
	public void saveEncrypted(int type, String text, File file) throws IOException {
		// Check for nullity.
		if(file != null) {
			// Our writer.
			BufferedWriter writer;
			
			// Check for the type of encryption.
			if(type == Task.URU) {
				// Open up a BufferedWriter over a long stream of writers and streams,
				// one of them my own!!
				writer = new BufferedWriter(new OutputStreamWriter(new WDYSOutputStream(new FileOutputStream(file), (int) text.length()), "US-ASCII"));
			}
			else if(type == Task.EOA) {
				// Open up a BufferedWriter over a long stream of writers and streams,
				// one of them my own!!
				writer = new BufferedWriter(new OutputStreamWriter(new WDYSVOutputStream(new FileOutputStream(file), (int) text.length()), "US-ASCII"));
			}
			else  throw new InternalError("Cannot save the encrypted file: Unsupported type.");
			
			// Write the string to the file.
			writer.write(text);
			
			// Close the writer.
			writer.close();
		}
	}
	
	/**
	 * Saves the given text into the given file.
	 * 
	 * @param text The text to save.
	 * @param file The file to save to.
	 * @throws IOException When streamstuffs go wrong.
	 */
	public void saveDecrypted(String text, File file) throws IOException {
		// Check for nullity.
		if(file != null) {
			// Open up a BufferedWriter over a long stream of writers and streams XD
			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), "US-ASCII"));
			
			// Write the tring to the file.
			writer.write(text);
			
			// Close the writer.
			writer.close();
		}
	}
	
	
	/**
	 * Makes the window ask the user for a file and returns that.
	 * 
	 * @return The selected file.
	 */
	protected File browse(DisplayFrame frame) {
		// Ask the window.
		return frame.askForFileSelection("Select file...", FileDialog.SAVE);
	}
	
	
	/**
	 * <p>This method opens a new window on the GUI, that contains the contents of the
	 * given file for editing. The contents are decrypted using the given type of
	 * encryption depending on the mode.</p>
	 * 
	 * <p>If the mode is to decrypt, the file will be decrypted before it is shown.
	 * However if the mode is to encrypt, the file will be shown as is, because one can
	 * assume that is what the user wants to see.</p>
	 * 
	 * @param file The file to show.
	 * @param type The type of encryption (Uru ot MYST V).
	 * @param mode The mode of encryption (encrypt or decrypt).
	 * @throws IOException When the window can't do its streamstuff without errors.
	 */
	public void openWindow(File file, int type, int mode) throws IOException {
		// Our reader.
		BufferedReader reader;
		
		// How much text there is going to be.
		int length;
		
		// Check the mode.
		if(mode == Task.ENCRYPT) {
			// The file is a normal text file.
			reader = new BufferedReader(new InputStreamReader(new FileInputStream(file), "US-ASCII"));
			
			// Set the length.
			length = (int) file.length();
		}
		else if(mode == Task.DECRYPT) {
			// The file is an encrypted file. Check it's type.
			if(type == Task.URU) {
				// Construct a Uru decrypter over the file.
				WDYSInputStream input = new WDYSInputStream(new FileInputStream(file));
				
				// Retrieve and set the length.
				length = input.getLength();
				
				// Construct an ASCII reader over the stream.
				reader = new BufferedReader(new InputStreamReader(input, "US-ASCII"));
			}
			else if(type == Task.EOA) {
				// Construct an EOA decrypter over the file.
				WDYSVInputStream input = new WDYSVInputStream(new FileInputStream(file));
				
				// Retrieve and set the length.
				length = input.getLength();
				
				// Construct an ASCII reader over the stream.
				reader = new BufferedReader(new InputStreamReader(input, "US-ASCII"));
			}
			else  throw new InternalError("Cannot open file: Unsupported type.");
		}
		else  throw new InternalError("Cannot open file: Unsupported mode.");
		
		// Read until we drop. Or actually, we cheat because we know the length :P
		char[] buffer = new char[length];
		
		// Initialise...
		int left = length;
		
		// ...and read until we've got everything.
		while(left > 0) {
			// Read.
			int read = reader.read(buffer, length - left, left);
			
			// Decrease how much we still need to read.
			if(read > 0)  left -= read;
			else  throw new IOException("Cannot open file: Not enough data.");
		}
		
		// Close the reader.
		reader.close();
		
		// Create the new window with the text from the character buffer, and add it to
		// our list.
		Frame frame = new DisplayFrame(this, file.getName(), new String(buffer));
		windowlist.add(frame);
	}
	
	
	/**
	 * Closes all the windows controlled by this class.
	 */
	public void closeAllWindows() {
		// End of program, it would seem...
		Iterator iterator = windowlist.iterator();
		
		// Go through all the windows.
		while(iterator.hasNext()) {
			// Get the window.
			DisplayFrame frame = (DisplayFrame) iterator.next();
			
			// Close it.
			frame.setVisible(false);
			frame.dispose();
			
			// Remove it from the list.
			iterator.remove();
		}
	}
	
	
	/**
	 * The list with open windows.
	 */
	protected ArrayList windowlist;
}