/*
 * Copyright 2005 Berend "Kirk" Wouda
 * 
 * This file is part of WDYS'd V.
 * 
 * WDYS'd V is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WDYS'd V is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with WDYS'd V; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


//GUI control class.
package com.kirkwarez.wdysdv.control;

// Import AWT classes.
import java.awt.*;
import java.awt.event.*;

// Import IO.
import java.io.*;

// Import the file filters.
import com.kirkwarez.kirkpack.io.filter.file.ExtensionFileFilter;
import com.kirkwarez.kirkpack.io.eoa.wdysv.WDYSVFileFilter;
import com.kirkwarez.kirkpack.io.uru.wdys.WDYSFileFilter;

// Import the GUI.
import com.kirkwarez.wdysdv.gui.WVFrame;

// Import the system classes.
import com.kirkwarez.wdysdv.system.*;

// Import the batch classes.
import com.kirkwarez.wdysdv.batch.*;

/**
 * This class acts a control class that processes all the user's actions in the main
 * window.
 * 
 * @author Berend "Kirk" Wouda
 * @version 5.00
 * @since 5.00
 */
public class WVControl extends WindowAdapter implements ActionListener, ItemListener {
	/**
	 * Constructs a new control class related with the given GUI.
	 */
	public WVControl(WVFrame gui) {
		// Set the gui.
		this.gui = gui;
		
		// Create a new Batch.
		batch = new Batch();
		
		// Create a new Processor.
		processor = new Processor();
		
		// Create a new display control.
		displaycontrol = new DisplayControl();
	}
	
	
	/**
	 * Closes the GUI and exists the program.
	 * 
	 * @param event The event that happened.
	 */
	public void windowClosing(WindowEvent event) {
		// Retrieve the frame.
		WVFrame frame = (WVFrame) event.getWindow();
		
		// Close all the open windows first.
		displaycontrol.closeAllWindows();
		
		// Hide the Window.
		frame.setVisible(false);
		
		// Dump the Window.
		frame.dispose();
		
		// Exit!
		System.exit(0);
	}
	
	
	/**
	 * Processes a checkbox event.
	 * 
	 * @param event The event.
	 */
	public void itemStateChanged(ItemEvent event) {
		// Find where it happened.
		Checkbox box = (Checkbox) event.getSource();
		
		// Check which checkbox got selected.
		if((box.getLabel().equals("Encrypt") || box.getLabel().equals("Decrypt")) && gui.getInFile() != null && gui.getOutFile() == null) {
			// Suggest a filename.
			setOutFile(getSuggestedOutFile(gui.getInFile(), gui.getInFile().getParentFile()));
		}
	}
	
	
	/**
	 * Processes an event created by the GUI because of user interaction.
	 * 
	 * @param event The event that happened.
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent event) {
		// Retrieve the action command from the event. 
		String command = event.getActionCommand();
		
		// Catch any errors that may occur and want to stop program execution.
		try {
			// Process what the event/user wants us to do.
			if(command.equals("BROWSEIN")) {
				// Let the user browse for the file and set that as input file.
				setInFile(browseInFile());
			}
			else if(command.equals("BROWSEOUT")) {
				// Let the user browse for the file and set that as output file.
				setOutFile(browseOutFile());
			}
			else if(command.equals("BATCH")) {
				// Check whether to add a file or folder.
				if(folder())  addTasks(gui.getInFile(), gui.getOutFile());
				else  addTask(getNewTask());
				
				// Clear the task.
				clearTask();
			}
			else if(command.equals("OPEN")) {
				// Check whether to open the current file or all the files in the batch.
				if(single()) {
					// Check whether to open just a file or a folder.
					if(folder())  openFolder(gui.getInFile(), gui.getType(), gui.getMode());
					else  openFile(gui.getInFile(), gui.getType(), gui.getMode());
					
					// Clear the task.
					clearTask();
				}
				else  openBatch(batch);
			}
			else if(command.equals("PROCESS")) {
				// Check whether to process the task or the batch.
				if(single()) {
					// Check whether to process just a file or a folder.
					if(folder())  processTasks(gui.getInFile(), gui.getOutFile());
					else  processTask(getNewTask());
					
					// Clear the task.
					clearTask();
				}
				else  processBatch(batch);
			}
			else if(command.equals("REMOVE")) {
				// Retrieve the selected task's index.
				int index = gui.getSelectedTask();
				
				// Check if there was anything selected and remove it if so.
				if(index >= 0)  removeTask(index);
			}
			else if(command.equals("REMOVEALL")) {
				// KILL
				removeAllTasks();
			}
			else if(command.equals("SENDBACK")) {
				// Retrieve the selected task's index.
				int index = gui.getSelectedTask();
				
				// Check if there was anything selected and send it back if so.
				if(index >= 0)  returnTask(gui.getSelectedTask());
			}
			else {
				// This is bad.
			}
		}
		catch(WVException e) {
			// Alert the user.
			gui.popupError("An error occured during the program: " + e.getMessage());
		}
		catch(IOException e) {
			// Alert the user.
			gui.popupError("An error occured while interacting with a file: " + e.getMessage());
		}
		catch(Exception e) {
			// Alert the user.
			gui.popupError("Something totally unexpected happened: " + e.getMessage());
		}
		
		// Update the GUI list.
		updateBatchList(batch);
	}
	
	
	/**
	 * Sets the input file field in the GUI to the given file. Now with added smart
	 * behaviour (the app will try to set the right settings depending on the filename)!
	 * 
	 * @param in The file to set it to.
	 */
	public void setInFile(File in) {
		// Check whether we CAN set.
		if(in != null) {
			// Tell the GUI.
			gui.setInFile(in);
			
			// Take a gander at the path name. See if we can figure out the type of the
			// file.
			if(in.getPath().matches(".*[Uu][Rr][Uu].*"))  gui.setType(Task.URU);
			else if(in.getPath().matches(".*[Mm][Yy][Ss][Tt].?[Vv5].*"))  gui.setType(Task.EOA);
			
			// Take a gander again. See if we can figure out the mode of operation.
			if(in.getName().endsWith(".txt"))  gui.setMode(Task.ENCRYPT);
			else {
				// Filters!
				WDYSFileFilter wdysfilter = new WDYSFileFilter();
				WDYSVFileFilter wdysvfilter = new WDYSVFileFilter();
				
				// Check the file against both filters.
				if(wdysfilter.accept(in) || wdysvfilter.accept(in))  gui.setMode(Task.DECRYPT);
			}
			
			// Check whether the output field is empty.
			if(gui.getOutFile() == null) {
				// Set the suggesed output filename.
				setOutFile(getSuggestedOutFile(in, in.getParentFile()));
			}
		}
	}
	
	/**
	 * Sets the output file field in the GUI to the given file.
	 * 
	 * @param out The file to set it to.
	 */
	public void setOutFile(File out) {
		// Check whether we CAN set.
		if(out != null) {
			// Tell the GUI.
			gui.setOutFile(out);
		}
	}
	
	
	/**
	 * Adds the given task to the batchlist.
	 * 
	 * @param task The task to add.
	 */
	public void addTask(Task task) {
		// Add the task.
		batch.addTask(task);
	}
	
	/**
	 * Adds the given folder of files to the batchlist by making tasks of each, using the
	 * settings in the GUI and adapting the filenames.
	 * 
	 * @param infolder The folder of files to batch.
	 * @param outfolder The output folder where the batched files go to.
	 * @throws WVException When folders aren't folders.
	 */
	public void addTasks(File infolder, File outfolder) throws WVException {
		// Check for fakes.
		if(outfolder != null) {
			// Check if the out folder is actually a folder. The in folder is, because it
			// was checked during method selection.
			if(outfolder.isDirectory()) {
				// Get all the files.
				File[] files = getFiles(infolder);
				
				// Go though all the files.
				for(int index = 0; index < files.length; index++) {
					// The current input file.
					File infile = files[index];
					
					// The current output file.
					File outfile = getSuggestedOutFile(infile, outfolder);
					
					// Make and add the task.
					addTask(new Task(infile, outfile, gui.getType(), gui.getMode()));
				}
			}
			else  throw new WVException("When folder batching, please fill in a folder in the output field.");
		}
		else  throw new WVException("Not all of the file fields are filled in.");
	}
	
	
	/**
	 * Opens the given file in the GUI in a new window.
	 * 
	 * @param file The file to open.
	 * @throws IOException When the window can't do its streamstuff without errors.
	 */
	public void openFile(File file, int type, int mode) throws IOException {
		// Open a new window on the GUI.
		displaycontrol.openWindow(file, type, mode);
	}
	
	/**
	 * Opens all the files in the given folder in the GUI in new windows.
	 * 
	 * @param folder The folder with the files to open.
	 * @throws IOException When the windows can't do their streamstuff without errors.
	 */
	public void openFolder(File folder, int type, int mode) throws IOException {
		// Get all the files in the folder.
		File[] files = getFiles(folder);
		
		// Go though all the files.
		for(int index = 0; index < files.length; index++) {
			// The current input file.
			File file = files[index];
			
			// Open a new window on the GUI.
			displaycontrol.openWindow(file, type, mode);
		}
	}
	
	/**
	 * Opens all the files in the batch in the GUI in new windows.
	 * 
	 * @param batch The batch with the files to open.
	 * @throws IOException When the windows can't do their streamstuff without errors.
	 */
	public void openBatch(Batch batch) throws IOException {
		// Go through the batch...
		for(int index = 0; index < batch.getNumberOfTasks(); index++) {
			// Get the task.
			Task task = batch.getTask(index);
			
			// Open a new window on the GUI.
			displaycontrol.openWindow(task.getInputFile(), task.getType(), task.getMode());
		}
	}
	
	
	/**
	 * Processes a single task.
	 * 
	 * @param task The task to process.
	 * @throws IOException When the file stuff barfs.
	 */
	public void processTask(Task task) throws IOException {
		// Call the Processor!
		processor.process(task);
	}
	
	/**
	 * Processes a folder of tasks :^P
	 * 
	 * @param infolder The folder of files to batch.
	 * @param outfolder The output folder where the batched tasks go to.
	 * @throws IOException When the file stuff barfs.
	 * @throws WVException When folders aren't folders.
	 */
	public void processTasks(File infolder, File outfolder) throws IOException, WVException {
		// Check for fakes.
		if(outfolder != null) {
			// Check if the out folder is actually a folder. The in folder is, because it
			// was checked during method selection.
			if(outfolder.isDirectory()) {
				// Get all the files.
				File[] files = getFiles(infolder);
				
				// Go though all the files.
				for(int index = 0; index < files.length; index++) {
					// The current input file.
					File infile = files[index];
					
					// The current output file.
					File outfile = getSuggestedOutFile(infile, outfolder);
					
					// Make and add the task.
					processTask(new Task(infile, outfile, gui.getType(), gui.getMode()));
				}
			}
			else  throw new WVException("When folder batching, please fill in a folder in the output field.");
		}
		else  throw new WVException("Not all of the file fields are filled in.");
	}
	
	/**
	 * Processes the given batch.
	 * 
	 * @param batch The batch with tasks to process.
	 * @throws IOException When the file stuff barfs.
	 */
	public void processBatch(Batch batch) throws IOException {
		// Call the Processor!
		processor.process(batch);
	}
	
	
	/**
	 * Removes the <code>Task</code> on the given index from the batch.
	 * 
	 * @param index The index of the task.
	 */
	public void removeTask(int index) {
		// Remove the task from the batch.
		batch.removeTask(index);
	}
	
	/**
	 * Removes all <code>Task</code>s from the batch.
	 */
	public void removeAllTasks() {
		// Remove all the tasks from the batch.
		batch.removeAllTasks();
	}
	
	/**
	 * Removes the <code>Task</code> on the given index from the batch and put it in the
	 * GUI.
	 * 
	 * @param index The index of the task.
	 */
	public void returnTask(int index) {
		// Remove the task from the batch.
		setNewTask(batch.removeTask(index));
	}
	
	
	/**
	 * Returns whether the GUI is in single or in batch mode.
	 * 
	 * @return Whether to perform a given task directly or run the batch instead.
	 */
	protected boolean single() {
		// Return whether there is text in the input field. If there is, we assume the
		// user wants to do that file directly.
		return gui.getInFile() != null;
	}
	
	/**
	 * Returns whether the input field is a file or a folder.
	 * 
	 * @return Whether to use the given file or all the files in the given folder.
	 */
	protected boolean folder() {
		// Get the input and output files.
		File input = gui.getInFile();
		
		// Return whether the input file is a folder.
		return input != null && input.isDirectory();
	}
	
	
	/**
	 * Makes the GUI ask the user for an input file and returns that.
	 * 
	 * @return The selected input file.
	 */
	protected File browseInFile() {
		// Ask the GUI.
		return gui.askForFileSelection("Select input file...", FileDialog.LOAD);
	}
	
	/**
	 * Makes the GUI ask the user for an output file and returns that.
	 * 
	 * @return The selected output file.
	 */
	protected File browseOutFile() {
		// Ask the GUI.
		return gui.askForFileSelection("Select output file...", FileDialog.SAVE);
	}
	
	
	/**
	 * Sets the suggested output filename in the output filename field.
	 * 
	 * @param in The input filename to construct it from.
	 * @param folder The folder to use as output folder.
	 * @return The suggested filename, or null if none could be determined.
	 */
	protected File getSuggestedOutFile(File in, File folder) {
		// Cache the filename.
		String filename = in.getName();
		
		// Check which mode we are in.
		if(gui.getMode() == Task.ENCRYPT) {
			// Set a suggested filename for the output file.
			if(filename.endsWith(".txt"))  return new File(folder, filename.substring(0, filename.length() - 4));
			else  return null;
		}
		else if(gui.getMode() == Task.DECRYPT) {
			// Set a suggested filename for the output file.
			return new File(folder, filename + ".txt");
		}
		else  throw new InternalError("Cannot construct suggested output filename: Unsupported mode.");
	}
	
	
	/**
	 * Returns all the files in the given folder that comply with the current settings in
	 * the GUI.
	 * 
	 * @return That what I just said.
	 * @throws NullPointerException When the given folder is null.
	 */
	protected File[] getFiles(File folder) {
		// Our filter.
		FileFilter filter;
		
		// Check which mode.
		if(gui.getMode() == Task.ENCRYPT) {
			// The filter is a normal extension filter that filters text files. Those are
			// the only files we know that might be what people want to be encrypted, and
			// also the only files we know how to rename for output.
			filter = new ExtensionFileFilter("txt");
		}
		else if(gui.getMode() == Task.DECRYPT) {
			// Check what type.
			if(gui.getType() == Task.URU) {
				// Use an Uru filter.
				filter = new WDYSFileFilter();
			}
			else if(gui.getType() == Task.EOA) {
				// Use a MYST V filter.
				filter = new WDYSVFileFilter();
			}
			else  throw new InternalError("Cannot list files: Unsupported type.");
		}
		else  throw new InternalError("Cannot list files: Unsupported mode.");
		
		// Return all the files.
		return folder.listFiles(filter);
	}
	
	
	/**
	 * Retrieves data from the GUI and makes a <code>Task</code> out of it.
	 * 
	 * @return That <code>Task</code>.
	 * @throws WVException If the user didn't fill in a file field.
	 */
	protected Task getNewTask() throws WVException {
		// Catch illegal argument exceptions.
		try {
			// Retrieve all the data from the GUI and make a nice task out of it.
			return new Task(gui.getInFile(), gui.getOutFile(), gui.getType(), gui.getMode());
		}
		catch(IllegalArgumentException e) {
			// Oooh, bad user! Not filling in filenames :o
			throw new WVException("Not all of the file fields are filled in.");
		}
	}
	
	/**
	 * Takes the given <code>Task</code> and puts its data in the GUI.
	 * 
	 * @param task The <code>Task</code> to put up in the GUI.
	 */
	protected void setNewTask(Task task) {
		// Retrieve all the data from the task and put it in the GUI.
		gui.setInFile(task.getInputFile());
		gui.setOutFile(task.getOutputFile());
		gui.setType(task.getType());
		gui.setMode(task.getMode());
	}
	
	
	/**
	 * Updates the lsit in the GUI with the given batch.
	 * 
	 * @param batch The batch to update with.
	 */
	protected void updateBatchList(Batch batch) {
		// Create a list of strings for the GUI.
		String[] list = new String[batch.getNumberOfTasks()];
		
		// Go through the batch.
		for(int index = 0; index < batch.getNumberOfTasks(); index++) {
			// Set the string.
			list[index] = batch.getTask(index).toString();
		}
		
		// Give it to the GUI.
		gui.setBatchList(list);
	}
	
	
	/**
	 * Clears the GUI's task settings.
	 */
	protected void clearTask() {
		// Reset the GUI's task elements.
		gui.setInFile(null);
		gui.setOutFile(null);
		gui.setType(Task.URU);
		gui.setMode(Task.ENCRYPT);
	}
	
	
	/**
	 * The batch.
	 */
	protected Batch batch;
	
	/**
	 * Our personal batch processor.
	 */
	protected Processor processor;
	
	
	/**
	 * The control class for the extra windows.
	 */
	protected DisplayControl displaycontrol;
	
	
	/**
	 * The associated GUI.
	 */
	protected WVFrame gui;
}